/*
 * cleanup -- decode tcpdump -x output to recreate packet data
 *
 * original code by Mark Fullmer, Ohio State University
 * ipsummary and reassembly mods by Steve Romig, OSU, 1996/1997
 *
 * compile with 'cc -o cleanup cleanup.c'
 *
 * usage:
 *	to do anything useful, use an adequate snaplength in tcpdump.
 *
 *	tcpdump -q -x -r file [filters] | cleanup -r
 *
 *		will reassemble the data into correct order and print it.
 *		this is only meaningful if you restrict the packet stream 
 *		to one side of a tcp session.
 *
 *	tcpdump -q -x -r file [filters] | cleanup -i
 *
 *		will reprint the per-packet summary line from tcpdump, 
 *		followed by a tab and the filtered contents of the packet
 *
 *	tcpdump -q -x -r file [filters] | cleanup -f
 *
 *		will reprint the packet contents, showing just the 
 *		printable ascii characters, plus \r, \n, backspace and 
 *		delete.
 *
 *	tcpdump -q -x -r file [filters] | cleanup
 *
 *		will reprint the literal packet contents
 *
 * notes:
 *	this is kinda of silly, should be incorporated into tcpdump 
 *	for efficiency
 *
 *	cleanup -r doesn't complain about missing data segments
 *
 * $Id: cleanup.c,v 1.1 1997/05/27 20:58:49 romig Exp romig $
 * $Log: cleanup.c,v $
 * Revision 1.1  1997/05/27 20:58:49  romig
 * Initial revision
 *
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>

#define RBSIZE	8192		/* size of initial buffer, increment size */

char getnext();
extern char *optarg;
int debug;
unsigned long seq;		/*  */
unsigned long isn = 0;		/*  */
unsigned long last = 0;		/*  */
unsigned long datasize;		/*  */
unsigned char *buffer;		/* tcp reassembly buffer */
unsigned long buffersize = 0;	/* size of reassembly buffer */

main(argc, argv)
int argc;
char **argv; {
	unsigned char datain[2000];
	char c;
	int snaplen = 84;
	int filter = 0;
	int reassemble = 0;	/* whether to reassemble or not */
	int ipsummary = 0;	/* whether to show ip summary or not */

	debug = 0;
	while ((c = getopt(argc, argv, "d:s:tfri")) != -1)
		switch(c) { 
		case 'd':
			debug = atoi(optarg);
			break;
		case 's':
			snaplen = atoi(optarg);
			break;
		case 'f': 
			filter = 1;
			break;
	        case 'r':
			reassemble = 1;
			break;
	        case 'i':
			ipsummary = 1;
			break;
		} /* switch */

	if (reassemble && (filter || ipsummary)) {
		fprintf(stderr, "can't use -f or -i with -r\n");
		exit(1);
	}

	if (filter && ipsummary) {
		fprintf(stderr, "can't use -f with -i\n");
		exit(1);
	}

	if (reassemble) {
		buffer = (char *) malloc(RBSIZE); /* initial buffer is 8k */
		buffersize = RBSIZE;	/* ditto */
		if (buffer == NULL) {
			fprintf(stderr, "malloc for reassembly buffer failed\n");
			exit(1);
		}
	}

	fgets(datain, 2000, stdin);	
	while (!feof(stdin)) {
		if (debug) printf("%s", datain); /* time, port, etc */
		printit(datain, snaplen, filter, reassemble, ipsummary);
	}

	if (reassemble) {
		if (write(fileno(stdout), buffer, last) < 0) {
			perror("write failed");
		}
	}
} /* main */

printit(datain, snaplen, filter, reassemble, ipsummary)
unsigned char *datain;
int snaplen;
int filter;
int reassemble;
int ipsummary;
{
	int bytes = 0;
	unsigned char *save_buffer;
	unsigned char outbuf[2048];
	unsigned char *o, *p;
	unsigned char low, high;
	unsigned int byte, x;
	unsigned char c;
	unsigned char hbyte, lbyte;
	unsigned long plength;	/* size of this packet, total */
	int dataoffset;		/* offset from packet start to data portion */
	unsigned long offset;	/* offset from start of stream to this data */
	unsigned char *rover;	/* tmp pointer */


	if (ipsummary) {
		/* datain should contain the tcpdump summary line with
		   the time, src and dst ip and stuff - change \n to \t
		   and print it */
		while ((rover = index(datain, '\n')) != NULL) {
			*rover = '\t';
		}
		printf("%s", datain);
	}

	o = outbuf;
	while (!feof(stdin)) {
		fgets(datain, 132, stdin);
		if (datain[0] != 9) break; /* this isn't a packet line */
		p = datain;
		/* save this line to buf in ascii */
		for (x = 0; x < 16; ++ x) {
			while ((!isalnum(*p)) && (*p != NULL)) ++ p;
			high = *p++;
			while ((!isalnum(*p)) && (*p != NULL)) ++ p;
			low = *p++;
	
			if (high == NULL) break; 	
			if (low == NULL) break;
			/* convert low and high to decimal */
			if ((high >= '0') && (high <= '9')) high -= '0'; else
			if ((high >= 'a') && (high <= 'f')) high = high - 'a'  + 10;

			if ((low >= '0') && (low <= '9')) low -= '0'; else
			if ((low >= 'a') && (low <= 'f')) low = low - 'a'  + 10;

			byte = (high * 16) + low;
			/* byte &= 127; */
			++ bytes; 
			*o++ = byte;
			} /* end for */
		} /* while */
	o = outbuf;
	hbyte = *(o+2);
	lbyte = *(o+3);
	plength = (hbyte * 256) + lbyte;
	if (debug) printf("Packet length %ld\n", plength);

	/* Get sequence number */
	seq = (*(o+24) << 24) | (*(o+25) << 16) | (*(o+26) << 8) | (*(o+27));
	if (isn == 0) {
		isn = seq+1;	/* set isn if it isn't already */
	}

	/* get offset to tcp data */
	lbyte = *(o+32);
	lbyte &= 0xF0;
	lbyte >>= 4;
	dataoffset = lbyte; 
	if (debug) printf("Data offset %d\n", dataoffset);
	dataoffset *= 4; /* # of 32 bit words */
	dataoffset += 20; /* ip header */

	if (debug) printf("seq %lu len %lu\n", seq, plength - dataoffset);

	if (reassemble) {
		offset = seq - isn;
		datasize = plength - dataoffset;

		if (debug) printf("  %lu %lu offset %lu length %lu last %lu\n", seq, isn, offset, datasize, last);

		if (datasize > 0) {
			/* grow the buffer if we need to */
			if (offset + datasize > buffersize) {
				buffersize += RBSIZE;
				save_buffer = buffer;
				buffer = (char *) realloc(buffer, buffersize);
				if (buffer == NULL) {
					fprintf(stderr, "realloc of reassembly buffer failed at %d bytes\n", buffersize);

					/* try to write what we have */
					if (write(fileno(stdout), save_buffer, last) < 0) {
					  perror("write failed");
					}

					exit(1);
				}
			}

			/* 
			 * copy the data portion of the packet to the buffer.
			 * the idea here is that we just write stuff 
			 * willy-nilly to the buffer, hopefully filling in
			 * the whole thing, without worrying about the order
			 * in which we received the data.  duplicate packets 
			 * should take care of themselves, and corrupted data
			 * should be replaced when we see the re-xmitted
			 * packet later.
			 */
			bcopy((o+dataoffset), (buffer+offset), datasize);
			if (offset + datasize > last) last = offset + datasize;
			/* 
			 * fixme: keep track of the segments that are filled
			 * in so we can detect missing segments
			 */
		}
	} else {
		/* else we're doing vanilla print-the-contents */
		while ((dataoffset < plength) && (dataoffset < snaplen)) {
			c = *(o+dataoffset);
			if (ipsummary) {
				/* in ipsummary mode, the printed data should
				   be printable ascii, with \r and \n quoted
				   so the whole packet appears on 1 line of
				   output */
				if (((isascii((int)c)) && (c > 31) && 
				     (c < 127)) || (c == 9)) {
					putchar(c);
				} else if (c == 10) {
					fputs("\\n", stdout);
				} else if (c == 13) {
					fputs("\\r", stdout);
				}
			} else if (filter) {
				/* in filter mode, print all printable ascii
				   characters, plus \n and \r.  this does
				   funny things to escape sequences and the 
				   like.  life's rough. */
				if (((isascii((int)c)) && (c > 31) && 
				     (c < 127))
/*				    || ((c == 9) || (c == 10) || (c == 13))) {*/
				    || ((c == 9) || (c == 10))) {
					putchar(c);
				} else if (c == 13) {
				        putchar('\n');
				} else if (c == 8) {
					fputs("^H", stdout);
				} else if (c == 127) {
					fputs("^?", stdout);
				}
			} else {
				/* else print the literal data */
				putchar(c);
			}
			++dataoffset;
		}

		if (dataoffset >= snaplen) {
			if (reassemble) {
				fprintf(stderr, "packet truncated\n");
			} else {
				printf("<TRUNC>");
			}
		}

/*		if (filter || ipsummary)  */
		if (ipsummary) 
		  putchar('\n');

		fflush(stdout);
	}
}
