package cache;

use strict;

##############################################################################
# cache package - routines for pulling sessions from log files and 
# caching them (for faster access) and for extracting session contents
#
# exports get_cache_file_name, make_and_save, make_and_save_extract
#
# copyright (c) 1997 by Steve Romig and The Ohio State University
#
##############################################################################
# $Id: cache.pm,v 1.1 1997/05/27 21:03:47 romig Exp $
##############################################################################
# $Log: cache.pm,v $
# Revision 1.1  1997/05/27 21:03:47  romig
# Initial revision
#

#
# services that should be cached
#
%cache::cache_service = (
		  'telnet'	=> 1,
		  'smtp'	=> 1,
		  'ftp'		=> 1,
		  'rsh'		=> 1,
		  'rlogin'	=> 1,
		  'irc'		=> 1,
		  'x'		=> 1,
		  );	

%cache::extract_service = (
		  'telnet'	=> 1,
		  'smtp'	=> 1,
		  'pop'		=> 1,
		  'imap'	=> 1,
		  'ftp'		=> 1,
		  'rsh'		=> 1,
		  'rlogin'	=> 1,
		  'irc'		=> 1,
		  );	

#
# return true if any of a list of service names appear in %cache_service.
#
sub should_be_cached {
    my(@tags) = @_;
    my $tag;

    foreach $tag (@tags) {
	if (defined($cache::cache_service{$tag})) {
	    return(1);
	}
    }

    return(0);
}

#
# return true if any of a list of service names appear in %extract_service.
#
sub should_be_extracted {
    my(@tags) = @_;
    my $tag;

    foreach $tag (@tags) {
	if (defined($cache::extract_service{$tag})) {
	    return(1);
	}
    }

    return(0);
}

#
# given the source file (path), ip and ports, get the name of the 
# needed cache file, check to see whether it exists, and create it if 
# it does not.
#
# does *not* check to see whether this session *should* be cached (as
# indicated by %cache_services) since we assume that this is done at the
# caller when necessary.
#
sub get_cache_file_name {
    my($file, $src_ip, $src_port, $dst_ip, $dst_port, $force) = @_;
    my $just_the_file = &main::basename($file);

    my $cache_file = "$main::cache_dir/$just_the_file:tcp:$src_ip.$src_port:$dst_ip.$dst_port";

    if (! -f $cache_file || $force) {
	my $command = "$main::tcpdump -r $file -w $cache_file -s 1600 host $src_ip and port $src_port and host $dst_ip and port $dst_port";

	print "  creating cache file $cache_file\n";
	print "   via '$command'\n"
	    if $main::debug;

	system($command);
    }

    return($cache_file);
}

#
# read through a summary for a log file and cache all of the 
# "interesting" sessions.  by cache we mean that the packets for that
# session are filtered out of the original log into a new log in the
# cache directory, so that future readers can read from there instead 
# of weeding through the entire (possibly long) log file.
#
sub make_and_save {
    my($file, $force) = @_;
    my($line);

    #
    # go through the summary and cache all tcp sessions
    #
    foreach $line (&summary::get($file, '', '')) {
	my($tmp, $tmp, $ip_a, $port_a, $ip_b, $port_b, $proto, @tags) = 
	    &summary::parse($line);

	next if $proto ne 'tcp';
	next if ! &should_be_cached(@tags);

	print "  cache $ip_a:$port_a > $ip_b:$port_b\n";

	&get_cache_file_name($file, $ip_a, $port_a, $ip_b, $port_b, $force);
    }
}

#
# extract the contents of a given session.  ick.
#
sub extract_contents {
    my ($file, $proto, $ip_a, $port_a, $ip_b, $port_b, $cleaner, $which) = @_;
    my $just_the_file = &main::basename($file);
    my $command;
    my $direction = "";

    if ($which ne "") {
	$direction = "and src port $which";
    }
    
    $command = "$main::tcpdump -q -x -n -s 1600 -r $file $proto and host $ip_a and port $port_a and host $ip_b and port $port_b $direction | $cleaner > $main::extract_dir/$just_the_file:$ip_a:$port_a:$ip_b:$port_b:$which";

    print "$command\n"
        if $main::debug;		

    system($command);
}

#
# read through a log summary and extract all of the interesting sessions.
#
sub make_and_save_extract {
    my($file, $force) = @_;
    my @summary;
    my $line;

    #
    # go through the summary and extract all interesting tcp sessions
    #
    foreach $line (&summary::get($file)) {
	my($tmp, $tmp, $ip_a, $port_a, $ip_b, $port_b, $proto, @tags) = 
	    &summary::parse($line);

	next if $proto ne 'tcp';
	next if ! &should_be_extracted(@tags);

	&extract_contents($file, $proto, $ip_a, $port_a, $ip_b, $port_b, $main::cleanup_text, $port_a);
	&extract_contents($file, $proto, $ip_a, $port_a, $ip_b, $port_b, $main::cleanup_text, $port_b);
    }
}

1;
