package summary;

use strict;

##############################################################################
# summary package - routines for creating and reading tcpdump log 
# summaries.
#
# exports make_summary, make_and_save, get, delete, parse
#
# copyright (c) 1997 by Steve Romig and The Ohio State University
#
##############################################################################
# $Id: summary.pm,v 1.1 1997/05/27 21:03:47 romig Exp $
##############################################################################
# $Log: summary.pm,v $
# Revision 1.1  1997/05/27 21:03:47  romig
# Initial revision
#

#
# Yuck.  Convert the tcpdump -q listing of a log file into a standard 
# format.
#
# Current format is something like this:
#
# arp, icmp: as is
# tcp, udp:
#  packets timestamp      ip-a_port-a  ip-b_port-b protocol type   group
#  10      21:53:12.12345 1.2.3.4_1000 5.6.7.8_23  tcp      telnet login
#
# FIXME: should aggregate icmp traffic?  or should we?
# FIXME: should lump port probes?
# FIXME: should handle different log file formats
#
sub make_summary {
    my($file) = @_;
    my(%index, $i, @result, @count);
    my $line;
    my($a_host, $a_port, $b_host, $b_port);
    my($time, $ip_a, $tmp, $ip_b, $proto, $rest);
    my($to_server, $type, $group);

#    $main::debug = 2;
    my $command = "$main::tcpdump -q -n -r $file |";
    print "$command\n" if $main::debug;

    open(TCP, $command) || die "couldn't open $command\n";

    #
    # push everything onto the @result array, to keep it all in chrono 
    # order, then cycle through and add packet counts to tcp/udp 
    # sessions.
    #
    while (<TCP>) {
	chop;

	#
	# icmp echo request/response
	# fixme: if we collapse this, there should be a way to show 
	# the expanded list via tcpdump -q or something.
	#
	if (m/icmp: echo/) {
	    ($time, $ip_a, $tmp, $ip_b, $proto, $rest) = split;

	    chop($ip_b); # remove trailing:

	    if (defined($index{$proto, $ip_a, $ip_b})) {
		$count[$index{$proto, $ip_a, $ip_b}]++;
	    } elsif (defined($index{$proto, $ip_b, $ip_a})) {
		$count[$index{$proto, $ip_b, $ip_a}]++;
	    } else {
		$line = sprintf("%s  %15s_  %15s icmp ping\n", 
				$time, $ip_a, $ip_b);
		push(@result, $line);
		push(@count, 1);
		print "icmp echo: $line" if $main::debug >= 2;
		$index{$proto, $ip_a, $ip_b} = $#result;
	    }
	} 

	#
	# icmp proto unreachable
	# fixme: if we collapse this, there should be a way to show 
	# the expanded list via tcpdump -q or something.
	#
	elsif (m/icmp: .* unreachable/) {
	    ($time, $ip_a, $tmp, $ip_b, $rest) = split;
	    $proto = 'icmp-unreach';

	    chop($ip_b); # remove trailing:

	    if (defined($index{$proto, $ip_a, $ip_b})) {
		$count[$index{$proto, $ip_a, $ip_b}]++;
	    } elsif (defined($index{$proto, $ip_b, $ip_a})) {
		$count[$index{$proto, $ip_b, $ip_a}]++;
	    } else {
		$line = sprintf("%s  %15s_  %15s icmp protocol unreachable\n", 
				$time, $ip_a, $ip_b);
		push(@result, $line);
		push(@count, 1);
		print "icmp proto unreachable: $line" if $main::debug >= 2;
		$index{$proto, $ip_a, $ip_b} = $#result;
	    }
	} 

	#
	# packet fragments
	# fixme: should be able to expand list somehow
	#
	elsif (m/\d: \(frag /) {
	    ($time, $ip_a, $tmp, $ip_b, $rest) = split;

	    chop($ip_b); # remove trailing:

	    if (defined($index{'frag', $ip_a, $ip_b})) {
		$count[$index{'frag', $ip_a, $ip_b}]++;
	    } elsif (defined($index{'frag', $ip_b, $ip_a})) {
		$count[$index{'frag', $ip_b, $ip_a}]++;
	    } else {
		$line = sprintf("%s  %15s_  %15s ip fragments\n", 
				$time, $ip_a, $ip_b);
		push(@result, $line);
		push(@count, 1);
		print "ip frag: $line" if $main::debug >= 2;
		$index{'frag', $ip_a, $ip_b} = $#result;
	    }
	} 

	#
	# igmp
	# fixme: if we collapse this, there should be a way to show 
	# the expanded list via tcpdump -q or something.
	#
	elsif (m/igmp/) {
	    ($time, $ip_a, $tmp, $ip_b, $proto, $rest) = split;

	    chop($ip_b); # remove trailing:

	    if (defined($index{$proto, $ip_a, $ip_b})) {
		$count[$index{$proto, $ip_a, $ip_b}]++;
	    } elsif (defined($index{$proto, $ip_b, $ip_a})) {
		$count[$index{$proto, $ip_b, $ip_a}]++;
	    } else {
		$line = sprintf("%s  %15s_  %15s igmp\n", 
				$time, $ip_a, $ip_b);
		push(@result, $line);
		push(@count, 1);
		print "igmp: $line" if $main::debug >= 2;
		$index{$proto, $ip_a, $ip_b} = $#result;
	    }
	}

	#
	# other icmp - pass through
	#
	elsif (m/icmp/) {
	    push(@result, sprintf("%10d  %s\n", 1, $_));
	    push(@count, 1);
	    print "icmp other: ", $_, "\n"  if $main::debug >= 2;
	}

	#
	# arp - pass
	#
	elsif (m/arp/) {
	    push(@result, sprintf("%10d  %s\n", 1, $_));
	    push(@count, 1);
	    print "arp: ", $_, "\n" if $main::debug >= 2;
	}

	#
	# tcp/udp - collapse to single sessions (or flows)
	#
	elsif (m/tcp/ || m/udp/) {
	    my($time, $ip_a, $tmp, $ip_b, $proto, $rest) = split;

	    chop($ip_b); # remove trailing:

	    if (defined($index{$proto, $ip_a, $ip_b})) {
		$count[$index{$proto, $ip_a, $ip_b}]++;
	    } elsif (defined($index{$proto, $ip_b, $ip_a})) {
		$count[$index{$proto, $ip_b, $ip_a}]++;
	    } else {
		($a_host, $a_port) = &services::split_ip_port($ip_a);
		($b_host, $b_port) = &services::split_ip_port($ip_b);

		($to_server, $type, $group) = 
		    &services::get_tags($proto, $a_host, $a_port, 
					$b_host, $b_port);

		$line = sprintf("%s  %15s_%-5s  %15s_%-5s %s %s %s\n", 
				$time, $a_host, $a_port, $b_host, $b_port,
				$proto, $type, $group);

		push(@result, $line);
		push(@count, 1);
		$index{$proto, $ip_a, $ip_b} = $#result;
		print "tcp/udp: ", $_, "\n" if $main::debug >= 2;
	    }
	} 

	#
	# anything else - pass through
	#
	else {
	    push(@result, sprintf("%10d  %s\n", 1, $_));
	    push(@count, 1);
	    print "other: ", $_, "\n" if $main::debug >= 2;
	}	
    }

    close(TCP);

    %index = ();

    for ($i = 0; $i <= $#result; $i++) {
	$result[$i] = sprintf("%10d  %s", $count[$i], $result[$i]);
    }

    @count = ();
    return(@result);
}

#
# make_and_save makes a summary of the tcpdump log file and saves it
# to the cache directory, returning a list of the summary lines
#
sub make_and_save {
    my($file, $force) = @_;
    my @summary;
    my $just_the_file = &main::basename($file);
    my $summary_file = "$main::summary_dir/$just_the_file.summary";
    my $line;

    #
    # make the summary and write it out
    #
    if (! -f $summary_file || $force) {
	@summary = &make_summary($file);
	open(SUMMARY, ">$summary_file") || 
	    die "cannot create $summary_file ($!)\n";

	foreach $line (@summary) {
	    print SUMMARY $line;
	}

	close(SUMMARY);
    }

    @summary = ();
}

#
# interface for outside world to read summary info for a log.  
#
# check to see whether a summary file exists in the summary directory: 
#
# if it doesn't exits, call make_summary to make one, make it and save it 
# to a file in the summary directory.
#
# if it does exist, pipe the contents through include and exclude 
# filtering and return as an array of lines of text.
#
sub get {
    my($file, $e_include, $e_exclude) = @_;
    my $command;
    my @result;
    my $just_the_file = &main::basename($file);
    my $summary_file = "$main::summary_dir/$just_the_file.summary";

    if (! -f $summary_file) {
	&make_and_save($file);
    }

    $command = "$main::cat $summary_file | $e_include $e_exclude";
    print "$command\n" if $main::debug;

    open(CMD, $command) ||
	die "cannot open $command ($!)\n";
    while (<CMD>) {
	push(@result, $_);
    }
    close(CMD);

    return(@result);
}

#
# delete the summary file corresponding to the named log file, if 
# it exists.
#
sub delete {
    my($file) = @_;
    my $just_the_file = &main::basename($file);
    my $summary_file = "$main::summary_dir/$just_the_file.summary";

    if (-f $summary_file) {
	my $result = unlink $summary_file;
	if ($result != 1) {
	    warn "could not delete $summary_file ($!)\n";
	}
    }
}

#
# parse_line splits a line from a summary into its various fields.
#
# FIXME: only works for tcp and udp entries currently.
#
sub parse {
    my($line) = @_;
    my($port_a, $port_b);

    chomp($line);
    my($packets, $time, $ip_a, $ip_b, $proto, @tags) = split(' ', $line);
    ($ip_a, $port_a) = &services::split_ip_port($ip_a);
    ($ip_b, $port_b) = &services::split_ip_port($ip_b);

    return($packets, $time, $ip_a, $port_a, $ip_b, $port_b, $proto, @tags);
}

#
# create a standard reference to a tcp or udp session within a log file.
#
sub make_ref {
    my($file, $proto, $ip_a, $port_a, $ip_b, $port_b) = @_;

    return("$file|p t ${ip_a}_$port_a ${ip_b}_$port_b $proto");
}
1;



